from typing import Callable
import inspect
import numpy as np
from . import get_rmse_from_function_array
from .. import density_functions as df


def get_density_functions_array():
    return [
        getattr(df, elemento) for elemento in dir(df) if inspect.isfunction(getattr(df, elemento))
    ]


def select_best_density_function_from_array(
    distance,
    density,
    aperture_diameter_data,
    helicopter_speed_data,
    swath_width,
    density_functions,
    flow_rate_function,
):
    rmse = get_rmse_from_function_array(
        distance,
        density,
        aperture_diameter_data,
        helicopter_speed_data,
        swath_width,
        density_functions,
        flow_rate_function,
    )
    es_mejor_funcion = rmse == rmse.min()
    return density_functions[np.where(es_mejor_funcion)[0][0]]


def get_best_density_function(
    distance: np.array,
    density: np.array,
    aperture_diameter_data: float,
    helicopter_speed_data: float,
    swath_width: float,
    flow_rate_function: Callable,
) -> Callable:
    """
    Select density function with minimum RMSE among the functions defined in
        submodule nerd.density_functions
    :param distance: Perpendicular distance in meters (m) from flight path
    :param density: Density of bait in kilograms per square meter (kg/m^2)
    :param aperture_diameter_data: Diameter (mm) of the dispersion bucket
        aperture
    :param helicopter_speed_data: Speed (m/s) of the helicopter during
        dispersion of bait
    :param swath_width: Width (m) of dispersion swath
    :param flow_rate_function: Function of mass flow rate (kg/s) with respect to
        aperture diameter (mm)
    :return: Function for density (kg/m^2) profile with respect to perpendicular
        distance (m) to flight path, swath width (m), and scale factor
    """
    density_functions = get_density_functions_array()
    return select_best_density_function_from_array(
        distance,
        density,
        aperture_diameter_data,
        helicopter_speed_data,
        swath_width,
        density_functions,
        flow_rate_function,
    )



from .fit_flow_rate import fit_flow_rate  # noqa
from .get_swath_width import get_swath_width  # noqa
from .model import model  # noqa
from .rmse import *  # noqa
from .best_density_function import *  # noqa



from matplotlib import path
from nerd import solver
from nerd.io import select_parameters_by_index, create_df_list
from nerd.density_functions import uniform
from scipy.interpolate import griddata
from shapely import geometry
from tqdm import tqdm

import fiona
import matplotlib.pyplot as plt
import numpy as np


def slope_between_two_points(y2, y1, x2, x1):
    return safe_divition(y2 - y1, x2 - x1)


def orthogonal_slope(slope):
    return safe_divition(-1, slope)


def safe_divition(numerator, denominator):
    if denominator == 0:
        return np.inf
    return numerator / denominator


def cell_edges_slopes(x, y, node_index):
    start_slope = orthogonal_slope(
        slope_between_two_points(
            y[node_index + 1], y[node_index - 1], x[node_index + 1], x[node_index - 1]
        )
    )
    end_slope = orthogonal_slope(
        slope_between_two_points(y[node_index + 2], y[node_index], x[node_index + 2], x[node_index])
    )
    return start_slope, end_slope


def calculate_cell_y_limits(slope, limit_x, x_coord, y_coord):
    return slope * (limit_x - x_coord) + y_coord


def cell_y_coordinates(start_orthogonal_slope, end_orthogonal_slope, x_rect, x, y, node):
    start_y1 = calculate_cell_y_limits(start_orthogonal_slope, x_rect[0], x[node], y[node])
    start_y2 = calculate_cell_y_limits(start_orthogonal_slope, x_rect[1], x[node], y[node])
    end_y2 = calculate_cell_y_limits(end_orthogonal_slope, x_rect[2], x[node + 1], y[node + 1])
    end_y1 = calculate_cell_y_limits(end_orthogonal_slope, x_rect[3], x[node + 1], y[node + 1])
    return [start_y1, start_y2, end_y2, end_y1, start_y1]


def calculate_cell_x_limits(r, orthogonal_slope, x_coord):
    return r / np.sqrt(1 + orthogonal_slope**2) + x_coord


def cell_x_coordinates(r, start_orthogonal_slope, end_orthogonal_slope, x, node):
    start_x1 = calculate_cell_x_limits(r, start_orthogonal_slope, x[node])
    start_x2 = calculate_cell_x_limits(-r, start_orthogonal_slope, x[node])
    end_x1 = calculate_cell_x_limits(r, end_orthogonal_slope, x[node + 1])
    end_x2 = calculate_cell_x_limits(-r, end_orthogonal_slope, x[node + 1])
    return [start_x1, start_x2, end_x2, end_x1, start_x1]


def generate_cell_from_coordinates(x, y, node, stripe_width, spatial_resolution):
    r = stripe_width / 2
    start_orthogonal_slope, end_orthogonal_slope = cell_edges_slopes(x, y, node)
    x_rect = cell_x_coordinates(r, start_orthogonal_slope, end_orthogonal_slope, x, node)
    y_rect = cell_y_coordinates(start_orthogonal_slope, end_orthogonal_slope, x_rect, x, y, node)
    x_rect, y_rect = check_directions(x_rect, y_rect)
    return x_rect, y_rect


def calculate_cell_density_in_border(x_rect, y_rect, n_point):
    startX = np.linspace(x_rect[0], x_rect[1], n_point)
    startY = np.linspace(y_rect[0], y_rect[1], n_point)
    endX = np.linspace(x_rect[3], x_rect[2], n_point)
    endY = np.linspace(y_rect[3], y_rect[2], n_point)
    xx = np.append(startX, endX)
    yy = np.append(startY, endY)
    return xx, yy


def density_in_tile(x_rect, y_rect, density_profile, n_point):
    xx, yy = calculate_cell_density_in_border(x_rect, y_rect, n_point)
    mean_xx = np.mean(xx)
    mean_yy = np.mean(yy)
    return lambda xq, yq: griddata(
        np.array([xx - mean_xx, yy - mean_yy]).T,
        np.tile(density_profile, 2),
        (xq - mean_xx, yq - mean_yy),
    )


def is_inside_tile(x_rect, y_rect, points):
    polygon_tile = [[x_rect[i], y_rect[i]] for i in range(len(x_rect))]
    poly = path.Path(polygon_tile)
    return poly.contains_points(points)


def calculate_directions(x_rect, y_rect):
    u, v = generate_tile_direction_arrays(x_rect, y_rect)
    theta1 = sign_of_direction(u, v)
    return theta1


def generate_tile_direction_arrays(x_rect, y_rect):
    u = np.array([x_rect[1] - x_rect[0], y_rect[1] - y_rect[0]])
    v = np.array([x_rect[2] - x_rect[3], y_rect[2] - y_rect[3]])
    return u, v


def sign_of_direction(u, v):
    inner = np.inner(u, v)
    norms = np.linalg.norm(u) * np.linalg.norm(v)
    return np.arccos(np.clip(inner / norms, -1.0, 1.0))


def reorder_end_tile(x_rect, y_rect):
    tempx1 = x_rect[3]
    x_rect[3] = x_rect[2]
    x_rect[2] = tempx1
    tempy1 = y_rect[3]
    y_rect[3] = y_rect[2]
    y_rect[2] = tempy1
    return x_rect, y_rect


def check_directions(x_rect, y_rect):
    startangle = calculate_directions(x_rect, y_rect)
    if startangle > np.pi / 2:
        x_rect, y_rect = reorder_end_tile(x_rect, y_rect)
    return x_rect, y_rect


def generate_contours(x_grid, y_grid, total_density, *args):
    contour = plt.contourf(x_grid, y_grid, total_density, *args)
    return contour, dict(zip(contour.collections, contour.levels))


def create_contour_polygon_list(contour, contour_dict):
    # Original code in https://github.com/chrishavlin/learning_shapefiles/blob/master/src/contourf_to_shp.py
    PolyList = []
    for col in contour.collections:
        z = contour_dict[col]
        for contour_path in col.get_paths():
            for ncp, cp in enumerate(contour_path.to_polygons()):
                lons = cp[:, 0]
                lats = cp[:, 1]
                new_shape = geometry.Polygon([(i[0], i[1]) for i in zip(lons, lats)])
                if ncp == 0:
                    poly = new_shape
                else:
                    poly = poly.difference(new_shape)

                PolyList.append({"poly": poly, "props": {"z": z}})
    return PolyList


def export_contour_list_as_shapefile(PolyList, output_path):
    schema = {"geometry": "Polygon", "properties": {"z": "float"}}
    with fiona.collection(output_path, "w", "ESRI Shapefile", schema) as output:
        for poly_list in PolyList:
            output.write(
                {"properties": poly_list["props"], "geometry": geometry.mapping(poly_list["poly"])}
            )


def generate_grid_density(x_coordinates, y_coordinates, spatial_resolution):
    x = np.arange(min(x_coordinates), max(x_coordinates), spatial_resolution)
    y = np.arange(min(y_coordinates), max(y_coordinates), spatial_resolution)
    x_grid, y_grid = np.meshgrid(x, y)
    return x_grid, y_grid


class Tracks:
    def __init__(self, track_data):
        self.track_data = track_data

    @property
    def x_coordinates(self):
        return self.track_data["easting"].to_numpy()

    @property
    def y_coordinates(self):
        return self.track_data["northing"].to_numpy()

    @property
    def bucket_logger(self):
        return self.track_data["Logging_on"].to_numpy()

    @property
    def helicopter_speed(self):
        return self.track_data["Speed"].to_numpy()

    @property
    def n_data(self):
        return len(self.track_data)


def calculate_total_density(
    track_data,
    config_file,
    spatial_resolution,
    flow_rate_function,
):
    tracks = Tracks(track_data)
    x_grid, y_grid = generate_grid_density(
        tracks.x_coordinates, tracks.y_coordinates, spatial_resolution
    )
    df_list = create_df_list(config_file)
    datafiles_lenghts = np.cumsum([len(df) for df in df_list])
    n_file = 0
    aperture_diameter, swap_width, density_function = select_parameters_by_index(
        config_file, n_file
    )
    x_grid_ravel = np.ravel(x_grid)
    y_grid_ravel = np.ravel(y_grid)
    total_density = np.zeros_like(x_grid_ravel)
    n = int(np.floor(swap_width / spatial_resolution))
    array_for_density = np.linspace(-swap_width / 2, swap_width / 2, n)
    for i in tqdm(range(tracks.n_data - 2)):
        if tracks.bucket_logger[i] == 0:
            continue
        else:
            if i >= datafiles_lenghts[n_file]:
                n_file += 1
            aperture_diameter, swap_width, density_function = select_parameters_by_index(
                config_file, n_file
            )
            density_function_lambda = solver(
                aperture_diameter,
                tracks.helicopter_speed[i],
                swap_width,
                density_function,
                flow_rate_function,
            )
            density_array = density_function_lambda(array_for_density)
            x_rect, y_rect = generate_cell_from_coordinates(
                tracks.x_coordinates, tracks.y_coordinates, i, swap_width, spatial_resolution
            )
            inside_mask = is_inside_tile(x_rect, y_rect, np.array([x_grid_ravel, y_grid_ravel]).T)
            sub_grid_x = x_grid_ravel[inside_mask]
            sub_grid_y = y_grid_ravel[inside_mask]
            cell_density = density_in_tile(x_rect, y_rect, density_array, n)(sub_grid_x, sub_grid_y)
            total_density[inside_mask] = total_density[inside_mask] + cell_density
    total_density_grid = np.reshape(total_density, x_grid.shape)
    return x_grid, y_grid, total_density_grid


def generate_uniform_density_array(density_value, stripe_width, spatial_resolution):
    r = int(stripe_width / 2)
    n = int(np.floor(stripe_width / spatial_resolution))
    rr = np.linspace(-r, r, n)
    normal_density_array = uniform(rr, stripe_width, density_value)
    return normal_density_array, n


def density_contours_intervals(density_value, total_density):
    mask_zeros = total_density != 0
    total_density_masked = total_density[mask_zeros]
    return np.unique(
        [
            total_density_masked.min() / 2,
            density_value / 2,
            density_value * 0.95,
            density_value * 1.05,
            np.min([2 * density_value, np.max(total_density)]),
            np.max(total_density),
        ]
    )



from typing import Callable
import numpy as np
from .. import solver


def model(
    aperture_diameter: np.array,
    helicopter_speed: np.array,
    swath_width: float,
    density_function: Callable,
    flow_rate_function: Callable,
) -> np.array:
    """
    Calculate the density matrix directly below the helicopter (nadir) as a
        function of the aperture diameter and the helicopter speed
    :param aperture_diameter: Diameter (mm) of the dispersion bucket aperture
    :param helicopter_speed: Speed (m/s) of the helicopter during dispersion of
        bait
    :param swath_width: Width (m) of dispersion swath
    :param density_function: Function for density (kg/m^2) profile with respect
        to perpendicular distance (m) to flight path, swath width (m), and scale
        factor
    :param flow_rate_function: Function of mass flow rate (kg/s) with respect to
        aperture diameter (mm)
    :return: Density matrix directly below the helicopter (nadir) as a function
        of aperture_diameter and helicopter_speed
    """
    distance = 0
    assert distance == 0
    densidad = np.zeros([len(helicopter_speed), len(aperture_diameter)])
    for i_diametro, diametro in enumerate(aperture_diameter):
        for i_rapidez, rapidez in enumerate(helicopter_speed):
            get_density = solver(
                diametro, rapidez, swath_width, density_function, flow_rate_function
            )
            densidad[i_rapidez][i_diametro] = get_density(distance)
    return np.array(densidad)



import numpy as np
from .. import solver


def get_rmse_from_function_array(
    distance,
    density,
    aperture_diameter,
    helicopter_speed,
    swath_width,
    density_functions_array,
    flow_rate_function,
):
    rmse = []
    for funcion_densidad in density_functions_array:
        rmse_auxiliar = get_rmse(
            distance,
            density,
            aperture_diameter,
            helicopter_speed,
            swath_width,
            funcion_densidad,
            flow_rate_function,
        )
        rmse.append(rmse_auxiliar)
    return np.array(rmse)


def get_rmse(
    distance,
    density,
    aperture_diameter,
    helicopter_speed,
    swath_width,
    density_function,
    flow_rate_function,
):
    density_profile_function = solver(
        aperture_diameter, helicopter_speed, swath_width, density_function, flow_rate_function
    )
    estimated_density = density_profile_function(distance)
    rmse = np.sqrt(np.mean((estimated_density - density) ** 2))
    return rmse



import pandas as pd
import utm


def geo2utm(csv_filename):
    data = pd.read_csv(csv_filename)
    data["easting"], data["northing"], data["zone_number"], data["zone_letter"] = utm.from_latlon(
        data.Lat.values, data.Lon.values
    )
    return data



from typing import Callable
import numpy as np


def fit_flow_rate(aperture_diameters: np.array, flow_rates: np.array) -> Callable:
    """
    Fit quadratic model for flow rate (kg/s) as a function of aperture diameter (mm)
    :param aperture_diameters: Array of diameters of the bucket aperture in
        millimetres (mm)
    :param flow_rates: Array of mass flow rate of bait in kg per second (kg/s)
    :return: Function of mass flow rate with respect to aperture diameter
    """
    coeficientes = np.polyfit(aperture_diameters, flow_rates, 2)
    return np.poly1d(coeficientes)



import numpy as np


def get_swath_width(distance: np.array, density: np.array, alpha: float = 0.05) -> float:
    """
    Get swath width (m) from density profile using data points with density greater than zero
    :param distance: Perpendicular distance in meters (m) from flight path
    :param density: Density of bait in kilograms per square meter (kg/m^2)
    :param alpha: Proportion (0<alpha<1) of discarded data points (to avoid outliers)
    :return: Swath width (m)
    """
    distancias_con_cebo = distance[density > 0]
    return np.diff(np.quantile(distancias_con_cebo, [alpha / 2, 1 - alpha / 2]))[0]



import numpy as np


def uniform(distance: float, width: float, parameter: float) -> float:
    """
    Uniform distribution for density profiles
    :param distance: Distance from flight path
    :param width: Swath width
    :param parameter: Free parameter to be fitted
    :return:
    """
    es_dentro = np.abs(distance) < width / 2
    return np.double(es_dentro) * parameter


def triangular(distance: float, width: float, parameter: float) -> float:
    """
    Triangular distribution for density profiles
    :param distance: Distance from flight path
    :param width: Swath width
    :param parameter: Free parameter to be fitted
    :return:
    """
    slope = -2 * parameter / width
    es_dentro = np.abs(distance) < width / 2  # pragma: no mutate
    return (slope * np.abs(distance) + parameter) * np.double(es_dentro)


def normal(distance: float, width: float, parameter: float) -> float:
    """
    Normal distribution for density profiles
    :param distance: Distance from flight path
    :param width: Swath width
    :param parameter: Free parameter to be fitted
    :return:
    """
    standard_deviation = width / 4
    return (
        parameter
        / np.sqrt(2 * np.pi * standard_deviation**2)
        * np.exp(-(distance**2) / (2 * standard_deviation**2))
    )



from nerd.io.geo2utm import geo2utm
from nerd.calibration.fit_flow_rate import fit_flow_rate
import nerd.density_functions
import pandas as pd
import os

column_names = ["date", "time", "Lat", "Lon", "Speed", "heading", "Logging_on", "altitude"]
flux_calibation_colums = ["aperture_diameter", "flux"]


def tracmap2csv(tracmap_filename, csv_filename):
    tracmap_data = pd.read_csv(
        tracmap_filename, header=None, names=column_names, usecols=[i for i in range(1, 9)]
    )
    tracmap_data.to_csv(csv_filename, index=False)


def import_tracmap(tracmap_filename, csv_filename="input_data.csv"):
    tracmap2csv(tracmap_filename, csv_filename)
    return geo2utm(csv_filename)


def import_calibration_data(flux_filename):
    flux_data = pd.read_csv(
        flux_filename,
        header=None,
        skiprows=1,
        names=flux_calibation_colums,
        usecols=[i for i in range(0, 2)],
    )
    return fit_flow_rate(flux_data["aperture_diameter"].to_numpy(), flux_data["flux"].to_numpy())


def check_output_directory(output_path):
    if not os.path.exists(output_path):
        os.mkdir(output_path)


def import_multifile_tracmap(config_file, csv_filename):
    df_list = create_df_list(config_file)
    df_concat = pd.concat(df_list)
    output_path = config_file.get("output_path")
    check_output_directory(output_path)
    concatenated_tracmap_path = "{}/{}".format(output_path, csv_filename)
    df_concat.to_csv(concatenated_tracmap_path, index=False)
    return geo2utm(concatenated_tracmap_path)


def create_df_list(config_file):
    df_list = [
        pd.read_csv(
            resources["input_data_path"],
            header=None,
            names=column_names,
            usecols=[i for i in range(1, 9)],
        )
        for resources in config_file["resources"]
    ]
    return df_list


def select_parameters_by_index(config_file, n_file):
    aperture_diameter = config_file["resources"][n_file]["aperture_diameter"]
    swap_width = config_file["resources"][n_file]["swap_width"]
    density_function = select_density_function(config_file["resources"][n_file]["density_function"])
    return aperture_diameter, swap_width, density_function


def select_density_function(function_name):
    return getattr(nerd.density_functions, function_name)



from nerd.io.import_data import (
    import_calibration_data,
    import_multifile_tracmap,
    check_output_directory,
)
from nerd.mapping.tiling import (
    calculate_total_density,
    density_contours_intervals,
    generate_contours,
)
import geojsoncontour
import json
import pandas as pd


class Nerd:
    def __init__(self, config_file_path):
        self.config_json_type_option = "series"
        self.config_file = pd.read_json(config_file_path, typ=self.config_json_type_option)
        self.tracmap_data = import_multifile_tracmap(
            self.config_file, "input_concatenated_data.csv"
        )
        self.spatial_resolution = self.config_file.get("spatial_resolution")
        self.flow_rate_function = import_calibration_data(
            self.config_file.get("input_calibration_data")
        )

    def calculate_total_density(self):
        self.x_grid, self.y_grid, self.total_density = calculate_total_density(
            self.tracmap_data,
            self.config_file,
            self.spatial_resolution,
            self.flow_rate_function,
        )

    def export_results_geojson(self, target_density):
        self.calculated_levels = density_contours_intervals(target_density, self.total_density)
        contours, contours_dict = generate_contours(
            self.x_grid, self.y_grid, self.total_density, self.calculated_levels
        )
        geojson = geojsoncontour.contourf_to_geojson(contourf=contours, unit="m")
        geojson = json.loads(geojson)
        check_output_directory(self.config_file.get("output_path"))
        with open(
            "{}/nerd_geojson.json".format(self.config_file.get("output_path")), "w"
        ) as outfile:
            json.dump(geojson, outfile)



from setuptools import setup, find_packages

setup(
    name="geci-nerd",
    version="0.3.0",
    packages=find_packages(),
    install_requires=[
        "descartes",
        "fiona",
        "geojsoncontour",
        "matplotlib",
        "numpy",
        "pandas",
        "scipy",
        "shapely",
        "tqdm",
        "utm",
    ],
)



from typing import Callable
from scipy.integrate import quad
from scipy.optimize import fsolve
import numpy as np


def solver(
    aperture_diameter: float,
    helicopter_speed: float,
    swath_width: float,
    density_function: Callable,
    flow_rate_function: Callable,
) -> Callable:
    """
    Fit model for bait density as a function of perpendicular distance in meters
        (m) from flight path
    :param aperture_diameter: Diameter (mm) of the dispersion bucket aperture
    :param helicopter_speed: Speed (m/s) of the helicopter during dispersion of
        bait
    :param swath_width: Width (m) of dispersion swath
    :param density_function: Function for density (kg/m^2) profile with respect
        to perpendicular distance (m) to flight path, swath width (m), and scale
        factor
    :param flow_rate_function: Function of mass flow rate (kg/s) with respect to
        aperture diameter (mm)
    :return: Function for bait density (kg/m^2) profile with respect to
        perpendicular distance (m) from flight path
    """

    def mass_conservation(parametro_libre):
        def sigma(distance):
            return density_function(distance, swath_width, parametro_libre)

        integrations_limits = swath_width / 2
        integral = quad(sigma, -integrations_limits, integrations_limits)[0]
        return integral - flow_rate_function(aperture_diameter) / helicopter_speed

    starting_root = np.random.rand()
    parametro_ajustado = fsolve(mass_conservation, starting_root)[0]
    return lambda x: density_function(x, swath_width, parametro_ajustado)
